
#
# Base class for DateTime::NoLeap, DateTime::AllLeap, DateTime::Fixed360Day
#
class DateTimeLike

  module DateTimeLikeExtension
    
    # Calculate Year/Month/Day fron Julian day number
    #
    # @param jday [Integer]
    #
    # @return [Array] [year, month, day]
    def jday2date (jday)
      dpy = self::DPY
      dpm = self::DPM
      jday = jday.round
      year = jday/dpy
      doy  = jday - year*dpy + 1
      year = year - 4712
      month = 0
      day   = 0
      days  = 0
      (1..12).each do |m|
        if days + dpm[m] >= doy
          month = m
          day = doy - days 
          break
        end
        days += dpm[m]
      end
      return year, month, day
    end
  
    # Creates a DateTimeFixedDPY object 
    # from the given Julian day number and Hour/Minute/Second.
    # 
    # @param jday [Interger]
    # @param hour [Interger]
    # @param minute [Integer]
    # @param second [Float]
    # @param offset [Rational]
    #
    # @return [DateTimeFixedDPY]
    def jd (jday, hour = 0, minute = 0, second = 0, offset = 0)
      year, month, day = jday2date(jday)
      return self.new(year, month, day, hour, minute, second, offset)
    end
  
  end
    
  include Comparable
  
  # Creates a DateTime object denoting the given calendar date.
  #
  # @param year [Integer]
  # @param month [Integer]
  # @param day [Integer]
  # @param hour [Integer]
  # @param minute [Integer]
  # @param second [Integer]
  # @param offset [Integer]
  # @param start [Integer]
  #
  def initialize (year = -4712, month = 1, day = 1, hour = 0, minute = 0, second = 0.0, offset = 0, start = nil)
    @year   = year.to_i
    @month  = month.to_i
    @day    = day.to_i
    @hour   = hour.to_i
    @minute = minute.to_i
    @second = second
    @offset = offset
    check_valid_datetime()
  end
    
  def check_valid_datetime
    unless valid_date?
      raise "invalid date for #{self.class} [year: #{@year}, month: #{@month}, day: #{@day}]"
    end
    unless valid_time?
      raise "invalid time for #{self.class} [hour: #{@hour}, minute: #{@minute}, second: #{@second}]"
    end
  end

  def valid_time?
    begin
      DateTime.new(2001, 1, 1, @hour, @minute, @second)
      true
    rescue
      false
    end
  end

  private :check_valid_datetime, :valid_time?

  attr_reader :year, 
              :month, 
              :day, 
              :hour, 
              :minute, 
              :offset

  # Returns the second (0-59)
  #
  # @return [Rational]
  def second
    return @second.floor
  end

  # Returns the fractional part of the second.
  #
  # @return [Rational]
  def second_fraction
    return @second.to_r - @second.floor
  end

  alias mon month
  alias min minute
  alias mday day
  alias sec second
  alias sec_fraction second_fraction

  # Duplicates self and resets its offset.
  #
  # @param offset [Numeric]
  #
  # @return [DateTime]
  def new_offset (offset = 0)
    gmt  = jd + fraction - @offset + offset.to_r
    jday = gmt.floor
    fday = gmt - gmt.floor
    return self.class.jd(jday, 0, 0, 0, offset) + fday
  end

  # Returns Day of Year
  #
  # @return [Integer]
  def yday
    dpm = self.class::DPM
    doy = @day
    (@month-1).step(1,-1) do |m|
      doy += dpm[m]
    end    
    return doy
  end

  # Calculates Julian day number from date part.
  # Note that this method does not take into account the offset and time. 
  # If you need a Julian day number that takes the time into account, use #ajd.
  #
  # @return [Integer]
  def jd
    return (self.class::DPY) * (@year + 4712) + (yday - 1)
  end

  # Calculates astronomical Julian day number.
  #
  # @return [Numeric]
  def ajd
    return jd + fraction - @offset - 1.quo(2) 
  end

  # Returns the day of week (0-6, Sunday is zero).
  #
  # @return [Integer]
  def wday 
    return (jd + 1 ) % 7
  end

  # Returns the day of week (0-6, Sunday is zero).
  #
  # @return [Integer]
  def commercial
    jd0 = (self.class::DPY) * (@year + 4712)
    wday0 = (jd0 + 1 ) % 7
    cwyear = @year
    yday1  = yday - wday0
    cweek = yday / 7 + 1
    cwday = yday % 7 + 1
    return cwyear, cweek, cwday
  end

  private :commercial

  # Returns a date object pointing other days after self. 
  #
  # @param days [Numeric]
  #
  # @return [DateTimeFixedDPY]
  def + (days)
    days = days.to_r + fraction 
    jday = jd.floor + days.floor
    fday = (days - days.floor)*24
    hour = fday.floor
    fday = (fday - hour)*60
    min  = fday.floor
    sec  = (fday - min)*60
    return self.class.jd(jday, hour, min, sec, offset)  
  end

  # Returns the difference between the two dates 
  # if the other is a datetime object. 
  # If the other is a numeric value, 
  # returns a date object pointing other days before self. 
  def - (other_or_days)
    case other_or_days
    when Numeric
      return self + (-other_or_days)      
    when self.class
      return self.jd - other_or_days.jd
    else
      raise "other shoud be date-time object or numeric value"
    end
  end

  # Returns a date object pointing n months after self. 
  # 
  # @param n [Integer]
  #
  # @return [DateTimeFixedDPY]
  def >> (n)
    return add_months(n)
  end

  # Returns a date object pointing n months before self.
  # 
  # @param n [Integer]
  #
  # @return [DateTimeFixedDPY]
  def << (n)
    return add_months(-n)
  end

  # Returns a date object pointing n years after self. 
  #
  # @param n [Integer]
  #
  # @return [DateTimeFixedDPY]
  def next_year (n = 1)
    return self.class::new(@year+n, @month, @day, @hour, @minute, @second, @offset)
  end

  # Returns a date object pointing n years before self. 
  #
  # @param n [Integer]
  #
  # @return [DateTimeFixedDPY]
  def prev_year (n = 1)
    return self.class::new(@year-n, @month, @day, @hour, @minute, @second, @offset)
  end

  # Returns a date object pointing n months after self. 
  #
  # @param n [Integer]
  #
  # @return [DateTimeFixedDPY]
  def next_month (n = 1)
    return add_months(n)
  end

  # Returns a date object pointing n months before self. 
  #
  # @param n [Integer]
  #
  # @return [DateTimeFixedDPY]
  def prev_month (n = 1)
    return add_months(-n)
  end

  def add_months (months)
    months = (@month + months).to_i - 1
    years  = months / 12
    month  = months % 12 + 1
    return self.class::new(@year+years, month, @day, @hour, @minute, @second, @offset)
  end

  private :add_months

  # Returns a date object pointing n days after self. 
  #
  # @param n [Integer]
  #
  # @return [DateTimeFixedDPY]
  def next_day (n = 1)
    return add_days(n)
  end

  # Returns a date object pointing n days before self. 
  #
  # @param n [Integer]
  #
  # @return [DateTimeFixedDPY]
  def prev_day (n = 1)
    return add_days(-n)
  end

  def add_days (days)
    return self.class::jd(jd+days.to_i, @hour, @minute, @second, @offset)
  end

  private :add_days
  
  # Returns a datetime object denoting the following day.
  #
  # @return [DateTime]
  def succ
    return add_days(1)
  end

  alias next succ

  # Returns time fraction in day units.
  #
  # @return [Rational]
  def fraction ()
    return (60*(60*@hour + @minute) + @second.to_r).quo(86400)
  end

  def format_offset
    oh = @offset*24
    return [
      oh >= 0 ? "+" : "-", 
      "%02i" % oh.abs.to_i, 
      ":",
      "%02i" % ((oh.abs - oh.abs.to_i)*60).round
    ].join("")    
  end

  def format_microsec
    microsec = ((@second - @second.floor)*1000000).round 
    if microsec == 0
      ""
    else
      ".%06d" % microsec
    end
  end

  private :format_offset, :format_microsec

  # Returns a string in an ISO 8601 format.
  #
  def to_s
    format("%04d-%02d-%02dT%02d:%02d:%02d%s%s",
           @year, @month, @day, 
           @hour, @minute, @second.floor, 
           format_microsec, 
           format_offset)
  end

  # Returns the value as a string for inspection.
  #
  def inspect
    sec  = fraction*86400
    isec = sec.floor
    fsec = ((sec - isec)*1000000).round
    offs = (offset*86400).round
    format("#<%s: %s ((%dj,%ds,%dn),%+ds)>", self.class, to_s, jd, isec, fsec, offs)
  end

  # Compares the two dates and returns -1, zero, 1 or nil. 
  # The other should be a date object or a numeric value 
  # as an astronomical Julian day number.
  def <=> (other)
    return self.ajd <=> other.ajd
  end
  
  def compare_md (other)
    sday = self.day + self.fraction + self.offset
    oday = other.day + other.fraction + other.offset
    if self.month > other.month
      return 1
    elsif self.month < other.month
      return -1
    else
      if sday > oday
        return 1
      elsif sday < oday
        return -1
      else
        return 0
      end
    end
  end

  def compare_d (other)
    sday = self.day + self.fraction + self.offset
    oday = other.day + other.fraction + other.offset
    if sday > oday
      return 1
    elsif sday < oday
      return -1
    else
      return 0
    end
  end

  # Calculate difference between the object and other object in years.
  #
  # @return [Integer]
  def difference_in_years (other)
    my    = self.new_offset(0)
    other = other.new_offset(0)
    return my.year - other.year + my.compare_md(other).quo(2)
  end
  
  # Calculate difference between the object and other object in months.
  #
  # @return [Integer]
  def difference_in_months (other)
    my    = self.new_offset(0)
    other = other.new_offset(0)
    return 12*(my.year - other.year) + my.month - other.month + my.compare_d(other).quo(2)  
  end
end

