
autoload :TZInfo, "tzinfo"

class TimeStep

  class Calendar

    using TimeStep::DateTimeExt

    # @private
    SYM_365_day = "365_day".intern
    # @private
    SYM_366_day = "366_day".intern
    # @private
    SYM_360_day = "360_day".intern

    DateTimeType = {
      :standard => DateTime,
      :proleptic_gregorian => DateTime,
      :proleptic_julian => DateTime,
      :noleap => DateTime::NoLeap,
      :allleap => DateTime::AllLeap,
      SYM_360_day => DateTime::Fixed360Day,
    }

    # Construct the object
    #
    def initialize (calendar = "standard", tz: nil)
      unless calendar.is_a?(String)
        raise ArgumentError, "argument calendar '#{calendar}' should be string" 
      end
      @name = calendar
      case @name.downcase.intern
      when :standard, :gregorian
        @calendar = :standard
      when :proleptic_gregorian
        @calendar = :proleptic_gregorian
      when :proleptic_julian, :julian
        @calendar = :proleptic_julian
      when :noleap, SYM_365_day
        @calendar = :noleap
      when :allleap, SYM_366_day
        @calendar = :allleap
      when SYM_360_day
        @calendar = SYM_360_day
      end
      if tz 
        raise "'standard' calendar needed for tz" unless @calendar == :standard
        case tz
        when nil
        when String
          @tz = TZInfo::Timezone.get(tz)
        when TZInfo::Timezone
          @tz = tz
        else
          raise "invalid tz specification"
        end
      end
    end
    
    attr_reader :name, :calendar, :tz
      
    def inspect 
      "#<TimeStep::Calendar calendar='#{@calendar.to_s}'>"
    end
    
    def == (other)
      return @calendar == other.calendar 
    end
        
    def valid_datetime_type? (time)
      return false unless time.is_a?(DateTimeType[@calendar])
      case @calendar
      when :standard
        return time.start == Date::ITALY
      when :proleptic_gregorian
        return time.start == Date::GREGORIAN
      when :proleptic_julian
        return time.start == Date::JULIAN
      else
        return true
      end 
    end
    
    # Parses the given representation of date and time in the calendar, 
    # and creates an date time instance.
    # 
    # @return [DateTime object]
    def parse (timespec, format: nil, offset: nil)
      return DateTime.parse_timestamp(timespec, calendar: @calendar.to_s, format: format, tz: @tz, offset: offset)
    end

    def time_new (year, month, day, hour=0, min=0, sec=0, offset=0)
      case @calendar
      when :standard
        time = DateTime.new(year, month, day, hour, min, sec, offset, Date::ITALY)
      when :proleptic_gregorian
        time = DateTime.new(year, month, day, hour, min, sec, offset, Date::GREGORIAN)
      when :proleptic_julian
        time = DateTime.new(year, month, day, hour, min, sec, offset, Date::JULIAN)
      when :noleap
        time = DateTime::NoLeap.new(year, month, day, hour, min, sec, offset)
      when :allleap
        time = DateTime::AllLeap.new(year, month, day, hour, min, sec, offset)
      when SYM_360_day
        time = DateTime::Fixed360Day.new(year, month, day, hour, min, sec, offset)
      end
      if @tz
        time = @tz.to_local(time)
      end
      return time
    end
  
    def jday2date (jday)
      case @calendar
      when :standard
        time = DateTime.jd(jday, 0, 0, 0, 0, Date::ITALY)
      when :proleptic_gregorian
        time = DateTime.jd(jday, 0, 0, 0, 0, Date::GREGORIAN)
      when :proleptic_julian
        time = DateTime.jd(jday, 0, 0, 0, 0, Date::JULIAN)
      when :noleap
        time = DateTime::NoLeap.jd(jday, 0, 0, 0, 0)
      when :allleap
        time = DateTime::AllLeap.jd(jday, 0, 0, 0, 0)
      when SYM_360_day
        time = DateTime::Fixed360Day.jd(jday, 0, 0, 0, 0)
      end
      if @tz
        time = @tz.to_local(time)
      end
      return time
    end

    def date2jday (year, month, day)
      return time_new(year, month, day).jd
    end
    
  end

end

