require "forwardable"

class TimeStep::Range 

  extend Forwardable

  def initialize (timestep, start = nil, last = nil, count: nil, ends: "[]")

    raise "'ends' option should be one of '[]', '()', '(]', '[)" unless ends =~ /\A[\[\(][\]\)]\z/
    include_start = ends[0] == "["
    include_last  = ends[1] == "]"

    if last
      case start
      when Numeric
      when DateTime, DateTimeLike, String
        start = timestep.index_at(start)
      else
        raise "unknown type of argument"
      end
      case last
      when Numeric
      when DateTime, DateTimeLike, String
        last   = timestep.index_at(last)
      else
        raise "unknown argument"
      end
    else
      case start
      when Integer
        count = start
        start = 0
        last  = count - 1
      when String
        raise "count should be set" unless count.is_a?(Integer)
        start = timestep.index_at(start)
        last  = start + count - 1
      when TimePeriod
        period = range
        return initialize(timestep, period.origin..period.last, ends: period.ends)
      else
        raise "unknown argument"
      end
    end

    if include_start
      @start = start.ceil
    else
      @start = start.floor + 1
    end
    
    if include_last
      @last = last.floor
    else
      @last = last.ceil - 1
    end

    @timestep = timestep.new_origin(timestep.time_at(@start))
    @count = @last - @start + 1
    
    @start_time = @timestep.time_at(@start)
    @last_time  = @timestep.time_at(@last)
  end

  attr_reader :timestep, :count, :start, :last, :start_time, :last_time

  def_delegators :@timestep, :numeric, :symbol, :interval, 
                             :origin, :offset, :calendar, :definition,
                             :parse, :time_at, :index_at, :duration_at, :[]
                             

  # FIXME
  def valid? (*indices)
    if @count
      if indices.size == 1
        index = indices.first
        if index >= 0 and index < @count 
          return true
        else
          return false
        end
      else
        return indices.map{|index| valid?(index) }
      end
    else
      if indices.size == 1
        return true
      else
        return indices.map{|index| true }
      end
    end
  end

  # Returns TimeStep object which has origin time determined 
  # by the given `index`.
  #
  # @param index [Numeric]
  #
  # @return [TimeStep]
  def shift_origin (index)
    case with
    when :index, "index"
      timestep = @timestep.shift_origin(index)
      return TimeStep::Range.new(timestep, count: @count)
    when :duration, "duration", :days, "days"
      timestep = @timestep.shift_origin(index, with: "duration")
      return TimeStep::Range.new(timestep, count: @count)
    end
  end

  # Returns TimeStep object which has origin time specified 
  # by the given `time`.
  #
  # @param time [DateTime]
  #
  # @return [TimeStep]
  def new_origin (time)
    timestep = @timestep.new_origin(time)
    return TimeStep::Range.new(timestep, count: @count)
  end

  include Enumerable
  
  def each_time (&block)
    if block
      @start.upto(@last) do |k|
        block.call(@timestep.time_at(k))
      end
    else
      return Enumerator.new {|y|
        @start.upto(@last) do |k|
          y << @timestep.time_at(k)
        end
      }
    end
  end

  alias each each_time
  
  def each_index (&block)
    (@start..@last).each(&block)
  end

  def map_index (&block)
    (@start..@last).map(&block)
  end

end

